"""
Here we dumped various functions which are not used in the 
current package. Well... Who knows? What if we will need them? 
Some of these functions are just old version of what we 
have now. 
"""


# rho4 represents the canonical homomorphism from B_4 to S_4
rho4 = (permut(3)(0,1), permut(3)(1,2), permut(2,3))
# the kernel of rho4 is PB_4


"""
Let us say that we are interested in tuples (w,m) which satisfy hexa1 only 
or hexa1 and hexa2 only or penta only 
"""
def gener_pr_hx1(t):
    """
    t represents a normal subgroup in PB_4 (which is normal in B_4)
    The function generates pairs (w,m) with gcd(2m+1,N0)==1 
    and satisfying ONLY hexa1
    """
    tPB3 = N_PB3(t); N0 = Nord(tPB3)
    for w in generWF2(tPB3[:2], timed=True):
        for m in range(N0):
            if gcd(2*m+1,N0)==1:
                if hexa1((w,m),tPB3):
                    yield (w,m)



def gener_pr_hx2(t):
    """
    t represents a normal subgroup in PB_4 (which is normal in B_4)
    The function generates pairs (w,m) with gcd(2m+1,N0)==1 
    and satisfying ONLY hexa1
    """
    tPB3 = N_PB3(t); N0 = Nord(tPB3)
    for w in generWF2(tPB3[:2], timed=True):
        for m in range(N0):
            if gcd(2*m+1,N0)==1:
                if hexa2((w,m),tPB3):
                    yield (w,m)
                    

def gener_pr_hx(t):
    """
    t represents a normal subgroup in PB_4 (which is normal in B_4)
    The function generates pairs (w,m) with gcd(2m+1,N0)==1 
    and satisfying ONLY hexa1 and hexa2 
    """
    tPB3 = N_PB3(t); N0 = Nord(tPB3)
    for w in generWF2(tPB3[:2], timed=True):
        for m in range(N0):
            if gcd(2*m+1,N0)==1:
                if hexa1((w,m),tPB3) and hexa2((w,m),tPB3):
                    yield (w,m)


def gener_pr_penta(t):
    """
    t represents an element of NFI_{PB_4}(B_4)
    The function generates pairs (w,0) which satisfy only the pentagon 
    relation.
    """
    tPB3 = N_PB3(t)
    for w in generWF2(tPB3[:2], timed=True):
        if penta((w,0),t):
            yield (w,0)
    

#tested
def in_PB4(t):
    """
    t represents a homomorphism phi from B_4 to S_n.
    The function returns True if the N =ker(phi) is in PB_4 
    """
    # first the index of N must be > or = 24 (the index of PB_4 in B_4)
    if PG(t).order()<24:
        return False
    tt = concat_tup(t,rho4)
    return PG(t).order() == PG(tt).order()

        
#tested  
def nest2perm(L):
    """
    L is a nested list of 1,2,...,d which represents a permutation 
    in S_d in the sense of cyclic decomposition. 
    The function returns the corresponding permutation as 
    an instance of the class Permutation 
    """
    tup = tuple(tuple(i-1 for i in a) for a in L)
    return permut(tup)
    
     

#Tested
def not_inlist_PB4(t, L):
    """
    t represents a normal subgroup in PB_4 and L is a list of 
    normal subgroups in PB_4. The function returns True if the normal 
    subgroup is NOT in the list. Otherwise False.
    """
    for tt in L:
        if sameNPB4(t,tt):
            return False
    return True
 

#tested
def sameNPB3(x,y):
    """
    x (resp. y) is a tuple of 3 permutations representing a homomorphism 
    from PB3 to S_n (resp to S_k). The function returns True if the kernels
    of the corresponding homomorphisms coincide. Otherwise False.
    """
    if x == y:
        return True
    indx = PG(x).order(); indy = PG(y).order()
    if indx != indy:
        return False
    z = concat_tup(x,y)
    return PG(z).order() == indx




#tested
def NPB3less_eq(x,y):
    """
    x,y represent homomorphisms from PB_3 to possibly distinct symmetric groups.
    Let N_x and N_y be kernels of these homomorphisms respectively.
    The function returns True of N_x <= N_y. Otherwise False.
    """
    if x==y:
        return True
    ind_x = PG(x).order(); ind_y = PG(y).order()
    if ind_x < ind_y:
        return False # a smaller subgroup must have a bigger index
    z = concat_tup(x,y)
    return PG(z).order()==ind_x



"""
A GT-pair (with k = 0) is represented by the pair sh = (w,m)
where w is a tuple of zeros and ones (w represents an element 
of the free group <x12,x23> in PB_3), m \in {0,1,..., Nord}
So sh corresponds to (f,m) where f = w(x12,x23) * c^0  

The old way of representing GT-pairs is (w,k,m).
Such a tuple (w,k,m) corresponds to (f,m) where 
f = w(x12,x23) * c^k 

Recall that c : = x23 x12 x13; c --> 1 0 2
"""

def to_ww(w,k):
    """
    w is a tuple of zeros and ones which represents an element of 
    the free group <x12,x23> and k is an integer. The function returns 
    a tuple of 0s 1s and 2s which represents an element in PB_3; 
    0 corresponds to x12, 1 corresponds to x23 and 2 corresponds to x13
    """
    ck=(1,0,2)*k
    return w + ck

#tested 
def tonew_sh(wm):
    """
    converts a representative of a GT-pair to the new format
    """
    return (wm[0], wm[2])

#tested
def hexa1_k(wm,tt):
    """
    OLD version of hexa1. k may be non-zero.
    wm is a candidate for a GT-pair and tt represents a normal subgroup in PB_3.
    The function returns True if the first hexagon is satisfied.
    Otherwise, False.
    """
    xy = tt[:2]; x12, x23 = xy
    c=cenPB3(tt); z = comp(x12, x23)**(-1); w,k,m = wm
    xz= (x12,z); yz = (x23,z) 
    fxy = compAll(tuple(xy[i] for i in w)) #the part without c
    fxz_inv = compAll(tuple(xz[i] for i in w))**(-1)
    fyz= compAll(tuple(yz[i] for i in w))
    tup = (x23**m, fxy, x12**m,fxz_inv, z**m, fyz,c**k)
    return is_id(compAll(tup))


def hexa2_k(wm, tt):
    """
    OLD version of hexa2. k may be non-zero.
    wm is a candidate for a GT-pair and tt represents a normal subgroup in PB_3.
    The function returns True if the second hexagon is satisfied. 
    Otherwise, False.
    """
    xy = tt[:2]; x12, x23 = xy
    c=cenPB3(tt); u = comp(x23, x12)**(-1); w,k,m = wm
    ux = (u,x12); uy = (u,x23)
    fxy_inv = compAll(tuple(xy[i] for i in w))**(-1)
    fux_inv = compAll(tuple(ux[i] for i in w))**(-1)
    fuy = compAll(tuple(uy[i] for i in w))
    tup = (fux_inv, x12**m, fxy_inv, x23**m, fuy, u**m,c**(-k))
    return is_id(compAll(tup))



def penta_k(wm, t):
    """
    OLD version of penta: k may be non-zero
    t is a tuple which represents a normal subgroup of PB_4
    and wm is a candidate for a shadow.
    The function returns True if sh satisfies the pentagon relation.
    Otherwise, False.
    """
    x12, x23, x13, x14, x24, x34 = t
    ww = to_ww(wm[0],wm[1]) # the PB_3-word
    x234=fi234(t)
    f234=compAll(tuple(x234[i] for i in ww))
    x123=fi123(t)
    f123=compAll(tuple(x123[i] for i in ww))
    x12_3_4=fi12_3_4(t)
    f12_3_4=compAll(tuple(x12_3_4[i] for i in ww))
    x1_23_4=fi1_23_4(t)
    f1_23_4=compAll(tuple(x1_23_4[i] for i in ww))
    x1_2_34=fi1_2_34(t)
    f1_2_34=compAll(tuple(x1_2_34[i] for i in ww))
    ###
    return compAll((f234,f1_23_4,f123))==comp(f1_2_34,f12_3_4)
    

"""
OTHER not very fast ways of generating charming GT-shadows are below...
"""


#tested
def generW(tt, timed=None):
    """
    a tuple (or list) tt represents a homomorphism from F_n to S_d. 
    G is image of this homomorphism.
    For every g in G the function "yields" exactly one representative in F_n
    (as a tuple of 0s,1s,...,n-1).
    Note that if G is big, then the generator will work for a while.
    """
    sz = PG(tt).order()#the order of the group
    d = tt[0].size # degree of the homomorphism (d in S_d)
    n = len(tt) # the number of the generators of the free group 
    if timed:
        T0 = time.time()
    G = [permut(d-1).rank()] # we will dump ranks of the new elements of G here
    W=[()]; L = [permut(d-1)] # L is the current layer of elements 
    #Note that compAll(()) = permut(0) and we need permut(d-1)
    yield () # the representative of the identity element
    while True:
        Wnew = []; Lnew = []
        for i in range(len(W)):
            for j in range(n):
                g = comp(L[i],tt[j]); rg = g.rank()
                if rg not in G:
                    Wnew.append(W[i] + (j,)); Lnew.append(g); G.append(rg)
        for w in Wnew:
            yield w
        W = Wnew[:]; L = Lnew[:]
        if timed:
            print((time.time()-T0)/60,' ',len(G))#reports time for every layer in minutes
        if len(G) == sz:
            break



#tested
def ww2w(h,ww):
    """
    ww is a word in F_n (without inverses) and 
    h represents a homomorphism from F_n to F_2;
    the function returns the image of ww in F_2
    """
    if ww ==():
        return ()
    out = ()
    for i in ww:
        out = out + h[i]
    return out


#tested
def elem_w(t,g):
    """
    t is a tuple which represents a group homomorphism from F_2 to S_d
    and g is an element in the image of this homomorphism. 
    The function returns a word in F_2 whose image is g
    WARNING: if g is not in the image of the homomorphism, the function
    will NOT stop working.
    """
    d = g.size
    if g == permut(d-1):
        return () # if g is the identity element
    for q in count(1):
        for w in prod((0,1),repeat=q):
            if compAll(tuple(t[i] for i in w))==g:
                return w

#tested
def generWH(t,H,timed=None):
    """
    t is a pair of permutations and H is a subgroup of 
    the permutation group generated by t[0] and t[1];
    for every h in H, the function 'yields' exactly one 
    representative in F_2 as a tuple of 0s and 1s
    """
    if timed:
        T0 = time.time()
    gens = H.strong_gens
    hh = tuple(elem_w(t,g) for g in gens) # we get a homomorphism from F_n to F_2
    print('I have just found representatives in F_2 for all strong generators.')
    if timed:
        print('It took me ',(time.time()-T0)/60,' minutes.')
    for ww in generW(gens,timed):
        yield ww2w(hh,ww)


#tested
def generW_awe(t, timed = None):
    """
    t is a tuple which represents a homomorphism from F_2 to S_d;
    for every g in the commutator subgroup of PG(t), the function 
    yields exactly one element representative in F_2
    """
    G = PG(t)
    for w in generWH(t, G.commutator(G,G) ,timed):
        yield w


#tested
def gener_GT_awe(t):
    """
    t represents an element N in NFI_{PB_4}(B_4).
    The function generates all charming GT-shadows
    corresponding to N. Now we only look at words whose
    image belongs to the commutator subgroup of the group
    generated by tPB3[:2]
    """
    tPB3 = N_PB3(t); N0 = Nord(tPB3)
    sz = PG(tPB3).order()
    szF2 = PG(tPB3[:2]).order()
    for w in generW_awe(tPB3[:2], timed=True):
        if penta((w,0),t):
            for m in range(N0):
                if gcd(2*m+1,N0)==1:
                    if hexa1((w,m),tPB3) and hexa2((w,m),tPB3):
                        src = sourcePB3((w,m),tPB3)
                        if PG(src).order()==sz and PG(src[:2]).order()==szF2:
                            yield (w,m)

"""
Next goes yet another way to generate charming GT-shadows.
"""
    

#tested
def elem_w1(t,g):
    """
    t is a tuple/list which represents a group homomorphism from F_n to S_d
    and g is an element in the image of this homomorphism. 
    The function returns the shortest word in F_n whose image is g
    WARNING: if g is not in the image of the homomorphism, the function
    will not stop working.
    """
    d = t[0].size # degree of the homomorphism (d in S_d)
    if g == permut(d-1):
        return ()
    G = [permut(d-1).rank()] # we will dump ranks of the new elements of G here
    W=[()]; L = [permut(d-1)] # L is the current layer of elements 
    while True:
        Wnew = []; Lnew = []
        for i in range(len(W)):
            for j in range(len(t)):
                h = comp(L[i],t[j])
                if h == g:
                    return W[i] + (j,)
                rh = h.rank()
                if rh not in G:
                    Wnew.append(W[i] + (j,)); Lnew.append(h); G.append(rh)
        W = Wnew[:]; L = Lnew[:]


#tested
def generWH1(t,H,timed=None):
    """
    t is a pair of permutations and H is a subgroup of 
    the permutation group generated by t[0] and t[1];
    for every h in H, the function 'yields' exactly one 
    representative in F_2 as a tuple of 0s and 1s
    """
    if timed:
        T0 = time.time()
    gens = H.strong_gens; hh = []
    for g in gens:
        print('I am working on the ', gens.index(g),'th strong generator.')
        hh.append(elem_w1(t,g)) 
    #we now have a homomorphism from F_n to F_2; it is represente by hh
    print('I have just found representatives in F_2 for all strong generators.')
    if timed:
        print('It took me ',(time.time()-T0)/60,' minutes.')
    for ww in generW(gens,timed):
        yield ww2w(hh,ww)


#tested
def generW_awe1(t, timed = None):
    """
    t is a tuple which represents a homomorphism from F_2 to S_d;
    for every g in the commutator subgroup of PG(t), the function 
    yields exactly one element representative in F_2
    """
    G = PG(t)
    for w in generWH1(t, G.commutator(G,G) ,timed):
        yield w

#tested
def gener_GT_awe1(t):
    """
    t represents an element N in NFI_{PB_4}(B_4).
    The function generates all charming GT-shadows
    corresponding to N. Now we only look at words whose
    image belongs to the commutator subgroup of the group
    generated by tPB3[:2]. The function is different from 
    gener_GT_awe in the sense that it uses a different generator for 
    words in F_2
    """
    tPB3 = N_PB3(t); N0 = Nord(tPB3)
    sz = PG(tPB3).order()
    szF2 = PG(tPB3[:2]).order()
    for w in generW_awe1(tPB3[:2], timed=True):
        if penta((w,0),t):
            for m in range(N0):
                if gcd(2*m+1,N0)==1:
                    if hexa1((w,m),tPB3) and hexa2((w,m),tPB3):
                        src = sourcePB3((w,m),tPB3)
                        if PG(src).order()==sz and PG(src[:2]).order()==szF2:
                            yield (w,m)
    

#tested   
def passport(c):
    """
    Returns the passport of the constellation c. The output is a 
    list which consists of 3 lists; one for each permutation in the
    constellation. Each list shows the corresponding parition of n.
    """
    g,gg = c
    ggg = compAll(c)**(-1)
    n = c[0].size
    dictionaries = (g.cycle_structure, gg.cycle_structure, ggg.cycle_structure)
    out = [(),(),()]
    for i in range(len(dictionaries)):
        for key in dictionaries[i]:
            for j in range(1, n+1):
                if key == j:
                    for k in range(dictionaries[i][key]):
                        out[i] = out[i] + (key,)
    return tuple(out)



#tested
def min_form(g,H):
    """
    g is an element of G and H is a subgroup of G.
    The function returns the minimal element of the form g.h
    """
    if g in H.elements:
        return one(H.degree)
    out = g
    for h in H.elements:
        p = comp(g,h)
        if p.rank()<out.rank():
            out = p
    return out

#tested
#No need. There is a faster method .coset_transversals(H)
def cosets(G,H):
    """
    H is a proper non-trivial subgroup of G.
    The function returns a list of representatives 
    of all cosets. Two different elements are from different 
    cosets 
    """
    n = H.degree
    out = []
    L = list(G.elements)
    r = 0
    for hh in H.elements:
        r = r+1
        if r%100==0:
            print(r) # to see how we are doing...
        L.remove(hh)
    print('The list is formed.')
    while len(L)>0:
        print(len(L)) #to see how we are doing...
        g = L[0]
        out.append(g)
        for h in H.elements:
            L.remove(comp(g,h))
    new = [min_form(gg,H) for gg in out]
    return [one(n)]+new





         
"""
G is a permutation group and H < = G. 
We need to describe the action of G on G/H
"""
#
def to_perm(g,L,G,H):
    """
    g is a permutation, L is the complete list of 
    coset representative G/H coming from G.coset_transversal(H)
    of H in G (g belongs to G).
    The function returns the corresponding permutation in S_d
    where d is the index of H in G.
    """
    out =[]
    for i in range(len(L)):
        p = G._coset_representative(comp(g,L[i]),H) 
        out.append(L.index(p))
    return permut(out)



"""
Let t be a tuple of tuples of the form ((0,1,..,k_1), (k_1+1,...k_2+1),...)
toPerm(t) returns the permutation with the corresponding cycle structure.
For example, toPerm( ((0,1),(2,3,4)) ) returns Permutation(0,1)(2,3,4) in S_5
"""
#Tested
def toPerm(t):
    aux = []
    for a in t:
        aux = aux + list(a[1:] + (a[0],))
    return permut(aux)


#Tested
def minP(t):
    """
    t is a tuple of permutations. minP(t) returns the "minimal" permutation
    with respect to the rank function
    """
    aux = tuple(x.rank() for x in t)
    return t[aux.index(min(aux))]


# No need for this function. 
# In SymPy, there is the command G.is_transitive for a permutation
# group G.    
# It may work faster than the one below. 
#Tested
def transit(x):
    """
    Let x be a tuple of permutations in S_n and G be the group generated by x.
    transit(x) returns True if G acts transitively on {0,1,... ,n-1}.
    Otherwise False.
    """
    n = x[0].size # all entries of x are in S_n
    y = set(a for a in x if a != one(n))
    prev = {0}; nexT = {0}
    while True:
        nexT = nexT.union({pr[0]^pr[1] for pr in prod(prev, y)})
        if nexT == prev:
            break
        else:
            prev = nexT
    return len(nexT) == n
    


#Tested
def normal(x):
    """
    Let x be a tuple of permutations and G be the subgroup generated by entries of x.
    normal(x) returns True if the stabilzer of 0 is trivial. 
    We assume that G acts transtitively on {0,1,..., n-1}.
    """
    G = PG(x); n = x[0].size
    for g in G.elements:
        if 0^g == 0 and g != one(n):
            return False
    return True


#tested
def charm_fake(E,EE,L,LL):
    """
    EE is a finer compatible equivalence relations than E;
    L (resp. LL) is the complete list of (charming) GT-shadows
    whose target is E (resp. EE);
    the function returns True if all GT-shadows in L survive into EE;
    Otherwise, False.
    """
    if EE.finer(E):
        pass
    else:
        print('Sorry! Your second argument is not finer than the first one.')
        return None
    Test =[]
    for T in LL:
        T_E=T.proj(E)
        if T_E not in Test:
            Test.append(T_E)
    #Test contains all projections of GT-shadows from LL "to E"
    #without repetitions
    if len(Test)<len(L):
        return False
    #do we have the same elements as in L?
    for T in Test:
        if T not in L:
            return False
    return True


# We did not use this function
#tested
def furusho_comm(E):
    """
    this function tests the onto version of Furusho phenomenon
    for a given equivalence relation E; Let G be the permutation group
    generated by the pair E.xy. Wpenta is the complete list of 
    words that represent elements of the commutator subgroup [G,G]
    and satify the pentagon relation with respect to E
    """
    Wpenta=[w for w in generWComm(E.xy, True) if penta((w,0),E.PB4)]
    print('Done with the list of words')
    print('Number of words satisfying the pentagon is ', len(Wpenta))
    d3=E.ind3()# we need it for the auxiliary function onto
    def onto(ww):
        for m in m_units(E.N0):
            tt = sourcePB3((ww,m), E.PB3)
            if PG(tt).order()==d3:
                return True
        return False
    W_onto=[w for w in Wpenta if onto(w)]
    print('The number of words satisfying onto property is ', len(W_onto))
    print('Please get ready!...')
    return [furusho(w,E) for w in W_onto]
    


